# ----------------------------------------------------------------------------
# Copyright (c) 2016-2022, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import os
import tempfile

from q2_types.feature_data import AlignedDNAFASTAFormat
from q2_types.tree import NewickFormat
from q2_phylogeny._raxml import run_command

_iqtree_defaults = {
    'seed': None,
    'n_cores': 1,
    'n_cores_max': None,
    'n_runs': 1,
    'substitution_model': 'MFP',
    'run_prefix': 'q2iqtree',
    'dtype': 'DNA',
    'n_init_pars_trees': None,
    'n_top_init_trees': None,
    'n_best_retain_trees': None,
    'n_iter': None,
    'stop_iter': None,
    'perturb_nni_strength': None,
    'spr_radius': None,
    'fast': False,
    'alrt': None,
    'abayes': False,
    'lbp': None,
    'allnni': False,
    'bnni': False,
    'safe': False,
    'bootstrap_replicates': 1000,
    'n_max_ufboot_iter': None,
    'n_ufboot_steps': None,
    'min_cor_ufboot': None,
    'ep_break_ufboot': None,
}


def _build_iqtree_command(
        alignment,
        seed: int = _iqtree_defaults['seed'],
        n_cores: int = _iqtree_defaults['n_cores'],
        n_cores_max: int = _iqtree_defaults['n_cores_max'],
        n_runs: int = _iqtree_defaults['n_runs'],
        substitution_model: str = _iqtree_defaults['substitution_model'],
        run_prefix: str = _iqtree_defaults['run_prefix'],
        dtype: str = _iqtree_defaults['dtype'],
        n_init_pars_trees: int = _iqtree_defaults['n_init_pars_trees'],
        n_top_init_trees: int = _iqtree_defaults['n_top_init_trees'],
        n_best_retain_trees: int = _iqtree_defaults['n_best_retain_trees'],
        n_iter: int = _iqtree_defaults['n_iter'],
        stop_iter: int = _iqtree_defaults['stop_iter'],
        perturb_nni_strength: float = _iqtree_defaults['perturb_nni_strength'],
        spr_radius: int = _iqtree_defaults['spr_radius'],
        allnni: bool = _iqtree_defaults['allnni'],
        fast: bool = _iqtree_defaults['fast'],
        alrt: int = _iqtree_defaults['alrt'],
        abayes: bool = _iqtree_defaults['abayes'],
        lbp: int = _iqtree_defaults['lbp'],
        safe: bool = _iqtree_defaults['safe']):

    cmd = ['iqtree']

    cmd += ['-st', str(dtype),
            '--runs', '%i' % n_runs,
            '-s', str(alignment),
            '-m', str(substitution_model),
            '-pre', str(run_prefix)]

    if n_cores == 'auto' and n_cores_max:
        cmd += ['-nt', 'AUTO', '--threads-max', '%i' % n_cores_max]
    elif n_cores == 'auto' and n_cores_max is None:
        cmd += ['-nt', 'AUTO']
    else:
        cmd += ['-nt', '%i' % n_cores]

    if seed:
        cmd += ['-seed', '%i' % seed]

    if safe:
        cmd += ['-safe']

    if fast:
        cmd += ['-fast']

    if alrt:
        cmd += ['-alrt', '%i' % alrt]

    if abayes:
        cmd += ['-abayes']

    if lbp:
        cmd += ['-lbp', '%i' % lbp]

    if allnni:
        cmd += ['-allnni']

    if n_init_pars_trees:
        cmd += ['-ninit', '%i' % n_init_pars_trees]

    if n_top_init_trees:
        cmd += ['-ntop', '%i' % n_top_init_trees]

    if n_best_retain_trees:
        cmd += ['-nbest', '%i' % n_best_retain_trees]

    if n_iter:
        cmd += ['-n', '%i' % n_iter]

    if stop_iter:
        cmd += ['-nstop', '%i' % stop_iter]

    if perturb_nni_strength:
        cmd += ['-pers', '%f' % perturb_nni_strength]

    if spr_radius:
        cmd += ['-sprrad', '%i' % spr_radius]

    return cmd


def iqtree(
    alignment: AlignedDNAFASTAFormat,
    seed: int = _iqtree_defaults['seed'],
    n_cores: int = _iqtree_defaults['n_cores'],
    n_cores_max: int = _iqtree_defaults['n_cores_max'],
    n_runs: int = _iqtree_defaults['n_runs'],
    substitution_model: str = _iqtree_defaults['substitution_model'],
    n_init_pars_trees: int = _iqtree_defaults['n_init_pars_trees'],
    n_top_init_trees: int = _iqtree_defaults['n_top_init_trees'],
    n_best_retain_trees: int = _iqtree_defaults['n_best_retain_trees'],
    n_iter: int = _iqtree_defaults['n_iter'],
    stop_iter: int = _iqtree_defaults['stop_iter'],
    perturb_nni_strength: float = _iqtree_defaults['perturb_nni_strength'],
    spr_radius: int = _iqtree_defaults['spr_radius'],
    allnni: bool = _iqtree_defaults['allnni'],
    fast: bool = _iqtree_defaults['fast'],
    alrt: int = _iqtree_defaults['alrt'],
    abayes: bool = _iqtree_defaults['abayes'],
    lbp: int = _iqtree_defaults['lbp'],
    safe: bool = _iqtree_defaults['safe'],
            ) -> NewickFormat:
    result = NewickFormat()

    with tempfile.TemporaryDirectory() as temp_dir:
        run_prefix = os.path.join(temp_dir, 'q2iqtree')
        cmd = _build_iqtree_command(alignment,
                                    seed=seed,
                                    n_cores=n_cores,
                                    n_cores_max=n_cores_max,
                                    n_runs=n_runs,
                                    substitution_model=substitution_model,
                                    run_prefix=run_prefix,
                                    n_init_pars_trees=n_init_pars_trees,
                                    n_top_init_trees=n_top_init_trees,
                                    n_best_retain_trees=n_best_retain_trees,
                                    n_iter=n_iter,
                                    stop_iter=stop_iter,
                                    perturb_nni_strength=perturb_nni_strength,
                                    spr_radius=spr_radius,
                                    allnni=allnni,
                                    fast=fast,
                                    alrt=alrt,
                                    abayes=abayes,
                                    lbp=lbp,
                                    safe=safe)
        run_command(cmd)

        tree_tmp_fp = os.path.join(temp_dir, '%s.treefile' % run_prefix)
        os.rename(tree_tmp_fp, str(result))

    return result


def _build_iqtree_ufbs_command(
        alignment,
        seed: int = _iqtree_defaults['seed'],
        n_cores: int = _iqtree_defaults['n_cores'],
        n_cores_max: int = _iqtree_defaults['n_cores_max'],
        n_runs:  int = _iqtree_defaults['n_runs'],
        substitution_model: str = _iqtree_defaults['substitution_model'],
        bootstrap_replicates: int = _iqtree_defaults['bootstrap_replicates'],
        run_prefix: str = _iqtree_defaults['run_prefix'],
        dtype: str = _iqtree_defaults['dtype'],
        n_init_pars_trees: int = _iqtree_defaults['n_init_pars_trees'],
        n_top_init_trees: int = _iqtree_defaults['n_top_init_trees'],
        n_best_retain_trees: int = _iqtree_defaults['n_best_retain_trees'],
        stop_iter: int = _iqtree_defaults['stop_iter'],
        perturb_nni_strength: float = _iqtree_defaults['perturb_nni_strength'],
        spr_radius: int = _iqtree_defaults['spr_radius'],
        n_max_ufboot_iter: int = _iqtree_defaults['n_max_ufboot_iter'],
        n_ufboot_steps: int = _iqtree_defaults['n_ufboot_steps'],
        min_cor_ufboot: float = _iqtree_defaults['min_cor_ufboot'],
        ep_break_ufboot: float = _iqtree_defaults['ep_break_ufboot'],
        allnni: bool = _iqtree_defaults['allnni'],
        alrt: int = _iqtree_defaults['alrt'],
        abayes: bool = _iqtree_defaults['abayes'],
        lbp: int = _iqtree_defaults['lbp'],
        bnni: bool = _iqtree_defaults['bnni'],
        safe: bool = _iqtree_defaults['safe']):
    # This is a separate command becuase there are several
    # bootstrap specific options.

    cmd = ['iqtree', '-bb', '%i' % bootstrap_replicates]

    cmd += ['-st', str(dtype),
            '--runs', '%i' % n_runs,
            '-s', str(alignment),
            '-m', str(substitution_model),
            '-pre', str(run_prefix)]

    if n_cores == 'auto' and n_cores_max:
        cmd += ['-nt', 'AUTO', '--threads-max', '%i' % n_cores_max]
    elif n_cores == 'auto' and n_cores_max is None:
        cmd += ['-nt', 'AUTO']
    else:
        cmd += ['-nt', '%i' % n_cores]

    if seed:
        cmd += ['-seed', '%i' % seed]

    if safe:
        cmd += ['-safe']

    if allnni:
        cmd += ['-allnni']

    if alrt:
        cmd += ['-alrt', '%i' % alrt]

    if abayes:
        cmd += ['-abayes']

    if lbp:
        cmd += ['-lbp', '%i' % lbp]

    if bnni:
        cmd += ['-bnni']

    if n_init_pars_trees:
        cmd += ['-ninit', '%i' % n_init_pars_trees]

    if n_top_init_trees:
        cmd += ['-ntop', '%i' % n_top_init_trees]

    if n_best_retain_trees:
        cmd += ['-nbest', '%i' % n_best_retain_trees]

    if stop_iter:
        cmd += ['-nstop', '%i' % stop_iter]

    if perturb_nni_strength:
        cmd += ['-pers', '%f' % perturb_nni_strength]

    if spr_radius:
        cmd += ['-sprrad', '%i' % spr_radius]

    if n_max_ufboot_iter:
        cmd += ['-nm', '%i' % n_max_ufboot_iter]

    if n_ufboot_steps:
        cmd += ['-nstep', '%i' % n_ufboot_steps]

    if min_cor_ufboot:
        cmd += ['-bcor', '%f' % min_cor_ufboot]

    if ep_break_ufboot:
        cmd += ['-beps', '%f' % ep_break_ufboot]

    return cmd


def iqtree_ultrafast_bootstrap(
    alignment: AlignedDNAFASTAFormat,
    seed: int = _iqtree_defaults['seed'],
    n_cores: int = _iqtree_defaults['n_cores'],
    n_cores_max: int = _iqtree_defaults['n_cores_max'],
    n_runs: int = _iqtree_defaults['n_runs'],
    substitution_model: str = _iqtree_defaults['substitution_model'],
    bootstrap_replicates: int = _iqtree_defaults['bootstrap_replicates'],
    n_init_pars_trees: int = _iqtree_defaults['n_init_pars_trees'],
    n_top_init_trees: int = _iqtree_defaults['n_top_init_trees'],
    n_best_retain_trees: int = _iqtree_defaults['n_best_retain_trees'],
    stop_iter: int = _iqtree_defaults['stop_iter'],
    perturb_nni_strength: float = _iqtree_defaults['perturb_nni_strength'],
    spr_radius: int = _iqtree_defaults['spr_radius'],
    n_max_ufboot_iter: int = _iqtree_defaults['n_max_ufboot_iter'],
    n_ufboot_steps: int = _iqtree_defaults['n_ufboot_steps'],
    min_cor_ufboot: float = _iqtree_defaults['min_cor_ufboot'],
    ep_break_ufboot: float = _iqtree_defaults['ep_break_ufboot'],
    allnni: bool = _iqtree_defaults['allnni'],
    alrt: int = _iqtree_defaults['alrt'],
    abayes: bool = _iqtree_defaults['abayes'],
    lbp: int = _iqtree_defaults['lbp'],
    bnni: bool = _iqtree_defaults['bnni'],
    safe: bool = _iqtree_defaults['safe']
                                ) -> NewickFormat:
    # NOTE: the IQ-TREE commands `-n` (called as `n_iter` in the `iqtree`
    # method) and `-fast` are not compatable with ultrafast_bootstrap `-bb`.
    result = NewickFormat()

    with tempfile.TemporaryDirectory() as temp_dir:
        run_prefix = os.path.join(temp_dir, 'q2iqtreeufboot')
        cmd = _build_iqtree_ufbs_command(
                    alignment,
                    seed=seed,
                    n_cores=n_cores,
                    n_cores_max=n_cores_max,
                    n_runs=n_runs,
                    substitution_model=substitution_model,
                    bootstrap_replicates=bootstrap_replicates,
                    run_prefix=run_prefix,
                    n_init_pars_trees=n_init_pars_trees,
                    n_top_init_trees=n_top_init_trees,
                    n_best_retain_trees=n_best_retain_trees,
                    stop_iter=stop_iter,
                    perturb_nni_strength=perturb_nni_strength,
                    spr_radius=spr_radius,
                    n_max_ufboot_iter=n_max_ufboot_iter,
                    n_ufboot_steps=n_ufboot_steps,
                    min_cor_ufboot=min_cor_ufboot,
                    ep_break_ufboot=ep_break_ufboot,
                    allnni=allnni,
                    alrt=alrt,
                    abayes=abayes,
                    lbp=lbp,
                    bnni=bnni,
                    safe=safe)
        run_command(cmd)
        tree_tmp_fp = os.path.join(temp_dir, '%s.treefile' % run_prefix)
        os.rename(tree_tmp_fp, str(result))

    return result



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2022, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import os
import tempfile
import subprocess

from random import randint

from q2_types.feature_data import AlignedDNAFASTAFormat
from q2_types.tree import NewickFormat

_raxml_versions = {
                   'Standard': '',
                   'SSE3': '-SSE3',
                   'AVX2': '-AVX2'
                   }


def run_command(cmd, verbose=True):
    if verbose:
        print("Running external command line application. This may print "
              "messages to stdout and/or stderr.")
        print("The command being run is below. This command cannot "
              "be manually re-run as it will depend on temporary files that "
              "no longer exist.")
        print("\nCommand:", end=' ')
        print(" ".join(cmd), end='\n\n')
    subprocess.run(cmd, check=True)


def _set_raxml_version(raxml_version='Standard', n_threads=1):
    if n_threads == 1:
        cmd = ['raxmlHPC' + _raxml_versions[raxml_version]]
        return cmd
    else:
        cmd = ['raxmlHPC-PTHREADS' + _raxml_versions[raxml_version],
               '-T %i' % n_threads]
        return cmd


def raxml(alignment: AlignedDNAFASTAFormat,
          seed: int = None,
          n_searches: int = 1,
          n_threads: int = 1,
          raxml_version: str = 'Standard',
          substitution_model: str = 'GTRGAMMA') -> NewickFormat:
    result = NewickFormat()

    cmd = _set_raxml_version(raxml_version=raxml_version, n_threads=n_threads)

    if seed is None:
        seed = randint(1000, 10000)

    runname = 'q2'
    with tempfile.TemporaryDirectory() as temp_dir:
        cmd += ['-m', str(substitution_model),
                '-p', str(seed),
                '-N', str(n_searches),
                '-s', str(alignment),
                '-w', temp_dir,
                '-n', runname]
        run_command(cmd)

        tree_tmp_fp = os.path.join(temp_dir, 'RAxML_bestTree.%s' % runname)
        os.rename(tree_tmp_fp, str(result))

    return result


def _build_rapid_bootstrap_command(alignment, seed, rapid_bootstrap_seed,
                                   bootstrap_replicates, substitution_model,
                                   temp_dir, runname):
    cmd = ['-f', 'a',  # always set, rapid bootstrapping
           '-m', str(substitution_model),
           '-p', str(seed),
           '-x', str(rapid_bootstrap_seed),
           '-N', str(bootstrap_replicates),
           '-s', str(alignment),
           '-w', temp_dir,
           '-n', runname]
    return cmd


def raxml_rapid_bootstrap(alignment: AlignedDNAFASTAFormat,
                          seed: int = None, rapid_bootstrap_seed: int = None,
                          bootstrap_replicates: int = 100, n_threads: int = 1,
                          raxml_version: str = 'Standard',
                          substitution_model: str = 'GTRGAMMA'
                          ) -> NewickFormat:
    result = NewickFormat()
    cmd = _set_raxml_version(raxml_version=raxml_version, n_threads=n_threads)

    if seed is None:
        seed = randint(1000, 10000)

    if rapid_bootstrap_seed is None:
        rapid_bootstrap_seed = randint(1000, 10000)

    runname = 'q2bootstrap'
    with tempfile.TemporaryDirectory() as temp_dir:
        cmd += _build_rapid_bootstrap_command(alignment, seed,
                                              rapid_bootstrap_seed,
                                              bootstrap_replicates,
                                              substitution_model, temp_dir,
                                              runname)
        run_command(cmd)

        tree_tmp_fp = os.path.join(temp_dir, 'RAxML_bipartitions.%s' % runname)
        os.rename(tree_tmp_fp, str(result))

    return result



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2022, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import biom
import numpy as np
import qiime2
import skbio


def filter_table(table: biom.Table, tree: skbio.TreeNode) -> biom.Table:
    """ Filter table to remove feature ids that are not tip ids in tree
    """
    tip_ids = set([t.name for t in tree.tips()])
    feature_ids = set(table.ids(axis='observation'))
    # ids_to_keep can only include ids that are in table
    ids_to_keep = tip_ids & feature_ids
    table.filter(ids_to_keep, axis='observation', inplace=True)
    return table


def filter_tree(tree: skbio.TreeNode,
                table: biom.Table = None,
                metadata: qiime2.Metadata = None,
                where: str = None,
                ) -> skbio.TreeNode:
    """
    Prunes a phylogenetic tree to match the input ids
    """
    # Checks the input metadata
    if ((table is None) & (metadata is None)):
        raise ValueError('A feature table, sequences or metadata must be '
                         'provided for filtering.')
    filter_refs = [table, metadata]
    if np.sum([(ref is not None) for ref in filter_refs]).sum() > 1:
        raise ValueError('Filtering can only be performed using one reference'
                         ' file. Please choose between filtering with a '
                         'feature table, sequences, or metadata.')
    if (where is not None) & (metadata is None):
        raise ValueError("Metadata must be provided if 'where' is specified")

    # Gets the list of IDs to keep
    if table is not None:
        ids_to_keep = table.ids(axis='observation')
    if metadata is not None:
        ids_to_keep = metadata.get_ids(where)

    # Gets the list of tips
    tip_ids = set([t.name for t in tree.tips()])

    # Checks for an intersection between ids
    if not set(tip_ids).issuperset(set(ids_to_keep)):
        raise ValueError('The ids for filtering must be a subset of '
                         'the tips in the tree.')
    sub_tree = tree.shear(ids_to_keep)
    sub_tree.prune()

    return sub_tree



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2022, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------


def align_to_tree_mafft_iqtree(ctx, sequences, n_threads=1,
                               mask_max_gap_frequency=1.0,
                               mask_min_conservation=0.40,
                               substitution_model='MFP',
                               fast=False, alrt=None,
                               seed=None, stop_iter=None,
                               perturb_nni_strength=None):
    mafft = ctx.get_action('alignment', 'mafft')
    mask = ctx.get_action('alignment', 'mask')
    iqtree = ctx.get_action('phylogeny', 'iqtree')
    midpoint_root = ctx.get_action('phylogeny', 'midpoint_root')

    aligned_seq, = mafft(sequences=sequences, n_threads=n_threads)
    masked_seq, = mask(alignment=aligned_seq,
                       max_gap_frequency=mask_max_gap_frequency,
                       min_conservation=mask_min_conservation)
    unrooted_tree, = iqtree(alignment=masked_seq, n_cores=n_threads,
                            fast=fast, alrt=alrt, seed=seed,
                            substitution_model=substitution_model,
                            stop_iter=stop_iter,
                            perturb_nni_strength=perturb_nni_strength)
    rooted_tree, = midpoint_root(tree=unrooted_tree)

    return (aligned_seq, masked_seq, unrooted_tree, rooted_tree)



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2022, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import os
import pkg_resources
import shutil
import unittest
import skbio
import tempfile

from qiime2.plugin.testing import TestPluginBase
from qiime2.util import redirected_stdio
from q2_types.feature_data import AlignedDNAFASTAFormat

from q2_phylogeny import raxml, raxml_rapid_bootstrap
from q2_phylogeny._raxml import (run_command, _build_rapid_bootstrap_command,
                                 _set_raxml_version)


class RaxmlTests(TestPluginBase):

    package = 'q2_phylogeny.tests'

    @classmethod
    def setUpClass(cls):
        super(TestPluginBase, cls).setUpClass()
        tmpdir = tempfile.mkdtemp()
        src = pkg_resources.resource_filename(cls.package, 'data')
        dst = os.path.join(tmpdir, 'data')
        shutil.copytree(src, dst)
        cls.data_dir = dst

    @classmethod
    def tearDownClass(cls):
        super(TestPluginBase, cls).setUpClass()
        shutil.rmtree(cls.data_dir)

    def get_data_path(self, filename):
        # Override TestPluginBase.get_data_path so that it returns paths to
        # temporary copies of test data.
        return os.path.join(self.data_dir, filename)

    def test_raxml(self):
        # Test that output tree is made.
        # Reads tree output and compares tip labels to expected labels.
        input_fp = self.get_data_path('aligned-dna-sequences-3.fasta')
        input_sequences = AlignedDNAFASTAFormat(input_fp, mode='r')
        with redirected_stdio(stderr=os.devnull):
            obs = raxml(input_sequences)
        obs_tree = skbio.TreeNode.read(str(obs))
        # load the resulting tree and test that it has the right number of
        # tips and the right tip ids
        tips = list(obs_tree.tips())
        tip_names = [t.name for t in tips]
        self.assertEqual(set(tip_names),
                         set(['GCA001510755', 'GCA001045515', 'GCA000454205',
                              'GCA000473545', 'GCA000196255', 'GCA000686145',
                              'GCA001950115', 'GCA001971985', 'GCA900007555']))

    def test_raxml_underscore_ids(self):
        # Test that output tree is made with underscores in tip IDs.
        # Some programs and python wrappers may strip underscores.
        # Reads tree output and compares tip labels to expected labels.
        input_fp = self.get_data_path('aligned-dna-sequences-4.fasta')
        input_sequences = AlignedDNAFASTAFormat(input_fp, mode='r')
        with redirected_stdio(stderr=os.devnull):
            obs = raxml(input_sequences)
        obs_tree = skbio.TreeNode.read(str(obs), convert_underscores=False)
        # load the resulting tree and test that it has the right number of
        # tips and the right tip ids
        tips = list(obs_tree.tips())
        tip_names = [t.name for t in tips]
        self.assertEqual(set(tip_names),
                         set(['GCA_001510755_1', 'GCA_001045515_1',
                              'GCA_000454205_1', 'GCA_000473545_1',
                              'GCA_000196255_1', 'GCA_002142615_1',
                              'GCA_000686145_1', 'GCA_001950115_1',
                              'GCA_001971985_1', 'GCA_900007555_1']))

    def test_set_raxml_version(self):
        obs_stand_1 = _set_raxml_version(raxml_version='Standard',
                                         n_threads=1)
        self.assertTrue('raxmlHPC' in str(obs_stand_1[0]))
        self.assertTrue(len(obs_stand_1) == 1)

        obs_sse3_1 = _set_raxml_version(raxml_version='SSE3', n_threads=1)
        self.assertTrue('raxmlHPC-SSE3' in str(obs_sse3_1[0]))
        self.assertTrue(len(obs_sse3_1) == 1)

        obs_avx2_1 = _set_raxml_version(raxml_version='AVX2', n_threads=1)
        self.assertTrue('raxmlHPC-AVX2' in str(obs_avx2_1[0]))
        self.assertTrue(len(obs_avx2_1) == 1)

        obs_stand_4 = _set_raxml_version(raxml_version='Standard',
                                         n_threads=4)
        self.assertTrue('raxmlHPC-PTHREADS' in str(obs_stand_4[0]))
        self.assertTrue('4' in str(obs_stand_4[1]))
        self.assertTrue(len(obs_stand_4) == 2)

        obs_sse3_4 = _set_raxml_version(raxml_version='SSE3', n_threads=4)
        self.assertTrue('raxmlHPC-PTHREADS-SSE3' in str(obs_sse3_4[0]))
        self.assertTrue('4' in str(obs_sse3_4[1]))
        self.assertTrue(len(obs_sse3_4) == 2)

        obs_avx2_4 = _set_raxml_version(raxml_version='AVX2', n_threads=4)
        self.assertTrue('raxmlHPC-PTHREADS-AVX2' in str(obs_avx2_4[0]))
        self.assertTrue('4' in str(obs_avx2_4[1]))
        self.assertTrue(len(obs_avx2_4) == 2)

    def test_raxml_version(self):
        # Test that an output tree is made when invoking threads.
        input_fp = self.get_data_path('aligned-dna-sequences-3.fasta')
        input_sequences = AlignedDNAFASTAFormat(input_fp, mode='r')

        with redirected_stdio(stderr=os.devnull):
            obs = raxml(input_sequences, raxml_version='SSE3')
        obs_tree = skbio.TreeNode.read(str(obs), convert_underscores=False)

        # load the resulting tree and test that it has the right number of
        # tips and the right tip ids
        tips = list(obs_tree.tips())
        tip_names = [t.name for t in tips]

        self.assertEqual(set(tip_names),
                         set(['GCA001510755', 'GCA001045515', 'GCA000454205',
                              'GCA000473545', 'GCA000196255', 'GCA000686145',
                              'GCA001950115', 'GCA001971985', 'GCA900007555']))

    def test_raxml_n_threads(self):
        # Test that an output tree is made when invoking threads.
        input_fp = self.get_data_path('aligned-dna-sequences-3.fasta')
        input_sequences = AlignedDNAFASTAFormat(input_fp, mode='r')

        with redirected_stdio(stderr=os.devnull):
            obs = raxml(input_sequences, n_threads=2)
        obs_tree = skbio.TreeNode.read(str(obs), convert_underscores=False)

        # load the resulting tree and test that it has the right number of
        # tips and the right tip ids
        tips = list(obs_tree.tips())
        tip_names = [t.name for t in tips]

        self.assertEqual(set(tip_names),
                         set(['GCA001510755', 'GCA001045515', 'GCA000454205',
                              'GCA000473545', 'GCA000196255', 'GCA000686145',
                              'GCA001950115', 'GCA001971985', 'GCA900007555']))

    def test_raxml_with_seed(self):
        # Test tip-to-tip dists are identical to manually run RAxML output.
        # This test is comparing an ordered series of tip-to-tip distances
        # to a tree output from a manual run of the default command:
        # raxmlHPC -m GTRGAMMA -p 1723 -s aligned-dna-sequences-3.fasta -n q2
        # NOTE: I cleanly rounded the tip-to-tip dists (i.e. `%.4f`) as RAxML
        # may return slightly different rounding errors on different
        # systems.
        input_fp = self.get_data_path('aligned-dna-sequences-3.fasta')
        input_sequences = AlignedDNAFASTAFormat(input_fp, mode='r')

        with redirected_stdio(stderr=os.devnull):
            obs = raxml(input_sequences, seed=1723)
        obs_tree = skbio.TreeNode.read(str(obs), convert_underscores=False)
        obs_tl = list(obs_tree.tip_tip_distances().to_series())
        obs_series = set(['%.4f' % e for e in obs_tl])

        exp_tree = skbio.TreeNode.read(self.get_data_path('test.tre'))
        exp_tl = list(exp_tree.tip_tip_distances().to_series())
        exp_series = set(['%.4f' % e for e in exp_tl])

        self.assertEqual(obs_series, exp_series)

    def test_raxml_model_choice(self):
        # Tip to tip dists should NOT be identical under different models.
        # Default is GTRGAMMA, we'll compare ouput to GRTGAMMAI & GTRCAT.
        # This test is comparing an ordered series of tip-to-tip distances.
        # Take note, that for this comparison to work, all must have the same
        # seed value set.
        input_fp = self.get_data_path('aligned-dna-sequences-3.fasta')
        input_sequences = AlignedDNAFASTAFormat(input_fp, mode='r')

        # default GTRGAMMA
        with redirected_stdio(stderr=os.devnull):
            gtrg = raxml(input_sequences, seed=1723)
            gtrg_tree = skbio.TreeNode.read(
                        str(gtrg), convert_underscores=False)
            gtrg_td = set(gtrg_tree.tip_tip_distances().to_series())

        # set GTRGAMMAI
        with redirected_stdio(stderr=os.devnull):
            gtrgi = raxml(input_sequences, seed=1723,
                          substitution_model='GTRGAMMAI')
            gtrgi_tree = skbio.TreeNode.read(
                         str(gtrgi), convert_underscores=False)
            gtrgi_td = set(gtrgi_tree.tip_tip_distances().to_series())

        # set GTRCAT
        with redirected_stdio(stderr=os.devnull):
            gtrcat = raxml(input_sequences, seed=1723,
                           substitution_model='GTRCAT')
            gtrcat_tree = skbio.TreeNode.read(
                          str(gtrcat), convert_underscores=False)
            gtrcat_td = set(gtrcat_tree.tip_tip_distances().to_series())

        # test pairs are not equivalent
        self.assertNotEqual(gtrg_td, gtrgi_td)
        self.assertNotEqual(gtrg_td, gtrcat_td)
        self.assertNotEqual(gtrgi_td, gtrcat_td)

    def test_raxml_num_searches(self):
        input_fp = self.get_data_path('aligned-dna-sequences-3.fasta')
        input_sequences = AlignedDNAFASTAFormat(input_fp, mode='r')
        with redirected_stdio(stderr=os.devnull):
            obs = raxml(input_sequences, seed=1723, n_searches=5)
        obs_tree = skbio.TreeNode.read(str(obs), convert_underscores=False)
        obs_tl = list(obs_tree.tip_tip_distances().to_series())
        obs_series = set(['%.4f' % e for e in obs_tl])

        exp_tree = skbio.TreeNode.read(self.get_data_path('test3.tre'))
        exp_tl = list(exp_tree.tip_tip_distances().to_series())
        exp_series = set(['%.4f' % e for e in exp_tl])
        self.assertEqual(obs_series, exp_series)

    def test_rapid_bootstrap_command(self):
        input_fp = self.get_data_path('aligned-dna-sequences-3.fasta')
        input_sequences = AlignedDNAFASTAFormat(input_fp, mode='r')
        with tempfile.TemporaryDirectory() as temp_dir:
            with redirected_stdio(stderr=os.devnull):
                obs = _build_rapid_bootstrap_command(input_sequences, 1723,
                                                     8752, 15, 'GTRGAMMA',
                                                     temp_dir, 'bs')
        self.assertTrue(str(input_sequences) in str(obs[11]))
        self.assertTrue('1723' in obs[5])
        self.assertTrue('8752' in obs[7])
        self.assertTrue('15' in obs[9])
        self.assertTrue('GTRGAMMA' in obs[3])
        self.assertTrue(str(temp_dir) in obs[13])
        self.assertTrue('bs' in obs[15])

    def test_raxml_rapid_bootstrap(self):
        # Test that output tree is made.
        # Reads tree output and compares tip labels to expected labels.
        input_fp = self.get_data_path('aligned-dna-sequences-3.fasta')
        input_sequences = AlignedDNAFASTAFormat(input_fp, mode='r')
        with redirected_stdio(stderr=os.devnull):
            obs = raxml_rapid_bootstrap(input_sequences)
        obs_tree = skbio.TreeNode.read(str(obs))
        # load the resulting tree and test that it has the right number of
        # tips and the right tip ids
        tips = list(obs_tree.tips())
        tip_names = [t.name for t in tips]
        self.assertEqual(set(tip_names),
                         set(['GCA001510755', 'GCA001045515', 'GCA000454205',
                              'GCA000473545', 'GCA000196255', 'GCA000686145',
                              'GCA001950115', 'GCA001971985', 'GCA900007555']))

    def test_raxml_rapid_bootstrap_n_threads(self):
        # Test that an output tree is made when invoking threads.
        input_fp = self.get_data_path('aligned-dna-sequences-3.fasta')
        input_sequences = AlignedDNAFASTAFormat(input_fp, mode='r')

        with redirected_stdio(stderr=os.devnull):
            obs = raxml_rapid_bootstrap(input_sequences, n_threads=2)
        obs_tree = skbio.TreeNode.read(str(obs), convert_underscores=False)

        # load the resulting tree and test that it has the right number of
        # tips and the right tip ids
        tips = list(obs_tree.tips())
        tip_names = [t.name for t in tips]

        self.assertEqual(set(tip_names),
                         set(['GCA001510755', 'GCA001045515', 'GCA000454205',
                              'GCA000473545', 'GCA000196255', 'GCA000686145',
                              'GCA001950115', 'GCA001971985', 'GCA900007555']))

    def test_raxml_rapid_bootstrap_with_seed(self):
        # Test tip-to-tip dists are identical to manually run RAxML output.
        # This test is comparing an ordered series of tip-to-tip distances
        # to a tree output from a manual run of the default command:
        #     raxmlHPC -f a -m GTRGAMMA -p 1723 -x 3871 -N 10
        #         -s aligned-dna-sequences-3.fasta -n q2
        # NOTE: I cleanly rounded the tip-to-tip dists (i.e. `%.4f`) as RAxML
        # may return slightly different rounding errors on different
        # systems (and at times, between conda environments).
        input_fp = self.get_data_path('aligned-dna-sequences-3.fasta')
        input_sequences = AlignedDNAFASTAFormat(input_fp, mode='r')

        # test that branchlengths are identical
        with redirected_stdio(stderr=os.devnull):
            obs = raxml_rapid_bootstrap(input_sequences, seed=1723,
                                        rapid_bootstrap_seed=3871,
                                        bootstrap_replicates=10)
        obs_tree = skbio.TreeNode.read(str(obs), convert_underscores=False)
        # sometimes we lose the last set of numbers on long floats
        obs_tl = list(obs_tree.tip_tip_distances().to_series())
        obs_series = set(['%.4f' % e for e in obs_tl])

        exp_tree = skbio.TreeNode.read(self.get_data_path('test2.tre'),
                                       convert_underscores=True)
        exp_tl = list(exp_tree.tip_tip_distances().to_series())
        exp_series = set(['%.4f' % e for e in exp_tl])
        self.assertEqual(obs_series, exp_series)

        # test that bootstrap supports are identical
        obs_bs = [node.name for node in obs_tree.non_tips()].sort()
        exp_bs = [node.name for node in exp_tree.non_tips()].sort()
        self.assertEqual(obs_bs, exp_bs)

    def test_run_not_verbose(self):
        input_fp = self.get_data_path('aligned-dna-sequences-3.fasta')
        input_sequences = AlignedDNAFASTAFormat(input_fp, mode='r')
        aligned_fp = str(input_sequences)

        with tempfile.TemporaryDirectory() as temp_dir:
            cmd = ['raxmlHPC',
                   '-m', 'GTRGAMMA',
                   '-p', '1723',
                   '-s', aligned_fp,
                   '-w', temp_dir,
                   '-n', 'q2']

            with redirected_stdio(stderr=os.devnull):
                run_command(cmd, verbose=False)
            obs_tree_fp = os.path.join(temp_dir, 'RAxML_bestTree.q2')
            obs_tree = skbio.TreeNode.read(str(obs_tree_fp),
                                           convert_underscores=False)
        # load the resulting tree and test that it has the right number of
        # tips and the right tip ids
        tips = list(obs_tree.tips())
        tip_names = [t.name for t in tips]
        self.assertEqual(set(tip_names),
                         set(['GCA001510755', 'GCA001045515',
                              'GCA000454205', 'GCA000473545',
                              'GCA000196255', 'GCA000686145',
                              'GCA001950115', 'GCA001971985',
                              'GCA900007555']))

    def test_run_rapid_bs_not_verbose(self):
        input_fp = self.get_data_path('aligned-dna-sequences-3.fasta')
        input_sequences = AlignedDNAFASTAFormat(input_fp, mode='r')
        aligned_fp = str(input_sequences)

        with tempfile.TemporaryDirectory() as temp_dir:
            cmd = ['raxmlHPC',
                   '-m', 'GTRGAMMA',
                   '-p', '1723',
                   '-s', aligned_fp,
                   '-w', temp_dir,
                   '-n', 'q2',
                   '-f', 'a',
                   '-x', '9834',
                   '-N', '10']

            with redirected_stdio(stderr=os.devnull):
                run_command(cmd, verbose=False)

            obs_tree_fp = os.path.join(temp_dir, 'RAxML_bipartitions.q2')
            obs_tree = skbio.TreeNode.read(str(obs_tree_fp),
                                           convert_underscores=False)
        # load the resulting tree and test that it has the right number of
        # tips and the right tip ids
        tips = list(obs_tree.tips())
        tip_names = [t.name for t in tips]
        self.assertEqual(set(tip_names),
                         set(['GCA001510755', 'GCA001045515',
                              'GCA000454205', 'GCA000473545',
                              'GCA000196255', 'GCA000686145',
                              'GCA001950115', 'GCA001971985',
                              'GCA900007555']))


if __name__ == "__main__":
    unittest.main()



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2022, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

from qiime2.plugin import (Plugin, Citations, Int, Range, Str, Choices, Bool,
                           Float, List, TypeMatch, Metadata)
from q2_types.tree import Phylogeny, Unrooted, Rooted
from q2_types.feature_data import FeatureData, AlignedSequence, Sequence
from q2_types.feature_table import (FeatureTable, Frequency,
                                    RelativeFrequency, PresenceAbsence,
                                    Composition)
from q2_types.distance_matrix import DistanceMatrix

import q2_phylogeny
import q2_phylogeny._examples as ex

_RAXML_MODEL_OPT = ['GTRGAMMA', 'GTRGAMMAI', 'GTRCAT', 'GTRCATI']
_RAXML_VERSION_OPT = ['Standard', 'SSE3', 'AVX2']
_IQTREE_DNA_MODELS = ['JC', 'JC+I', 'JC+G', 'JC+I+G', 'JC+R2', 'JC+R3',
                      'JC+R4', 'JC+R5', 'JC+R6', 'JC+R7', 'JC+R8', 'JC+R9',
                      'JC+R10', 'F81', 'F81+I', 'F81+G', 'F81+I+G', 'F81+R2',
                      'F81+R3', 'F81+R4', 'F81+R5', 'F81+R6', 'F81+R7',
                      'F81+R8', 'F81+R9', 'F81+R10', 'K80', 'K80+I', 'K80+G',
                      'K80+I+G', 'K80+R2', 'K80+R3', 'K80+R4', 'K80+R5',
                      'K80+R6', 'K80+R7', 'K80+R8', 'K80+R9', 'K80+R10',
                      'HKY', 'HKY+I', 'HKY+G', 'HKY+I+G', 'HKY+R2', 'HKY+R3',
                      'HKY+R4', 'HKY+R5', 'HKY+R6', 'HKY+R7', 'HKY+R8',
                      'HKY+R9', 'HKY+R10', 'TNe', 'TNe+I', 'TNe+G',
                      'TNe+I+G', 'TNe+R2', 'TNe+R3', 'TNe+R4', 'TNe+R5',
                      'TNe+R6', 'TNe+R7', 'TNe+R8', 'TNe+R9', 'TNe+R10', 'TN',
                      'TN+I', 'TN+G', 'TN+I+G', 'TN+R2', 'TN+R3', 'TN+R4',
                      'TN+R5', 'TN+R6', 'TN+R7', 'TN+R8', 'TN+R9', 'TN+R10',
                      'K81', 'K81+I', 'K81+G', 'K81+I+G', 'K81+R2', 'K81+R3',
                      'K81+R4', 'K81+R5', 'K81+R6', 'K81+R7', 'K81+R8',
                      'K81+R9', 'K81+R10', 'K81u', 'K81u+I', 'K81u+G',
                      'K81u+I+G', 'K81u+R2', 'K81u+R3', 'K81u+R4', 'K81u+R5',
                      'K81u+R6', 'K81u+R7', 'K81u+R8', 'K81u+R9', 'K81u+R10',
                      'TPM2', 'TPM2+I', 'TPM2+G', 'TPM2+I+G', 'TPM2+R2',
                      'TPM2+R3', 'TPM2+R4', 'TPM2+R5', 'TPM2+R6', 'TPM2+R7',
                      'TPM2+R8', 'TPM2+R9', 'TPM2+R10', 'TPM2u', 'TPM2u+I',
                      'TPM2u+G', 'TPM2u+I+G', 'TPM2u+R2', 'TPM2u+R3',
                      'TPM2u+R4', 'TPM2u+R5', 'TPM2u+R6', 'TPM2u+R7',
                      'TPM2u+R8', 'TPM2u+R9', 'TPM2u+R10', 'TPM3', 'TPM3+I',
                      'TPM3+G', 'TPM3+I+G', 'TPM3+R2', 'TPM3+R3', 'TPM3+R4',
                      'TPM3+R5', 'TPM3+R6', 'TPM3+R7', 'TPM3+R8', 'TPM3+R9',
                      'TPM3+R10', 'TPM3u', 'TPM3u+I', 'TPM3u+G', 'TPM3u+I+G',
                      'TPM3u+R2', 'TPM3u+R3', 'TPM3u+R4', 'TPM3u+R5',
                      'TPM3u+R6', 'TPM3u+R7', 'TPM3u+R8', 'TPM3u+R9',
                      'TPM3u+R10', 'TIMe', 'TIMe+I', 'TIMe+G', 'TIMe+I+G',
                      'TIMe+R2', 'TIMe+R3', 'TIMe+R4', 'TIMe+R5', 'TIMe+R6',
                      'TIMe+R7', 'TIMe+R8', 'TIMe+R9', 'TIMe+R10', 'TIM',
                      'TIM+I', 'TIM+G', 'TIM+I+G', 'TIM+R2', 'TIM+R3',
                      'TIM+R4', 'TIM+R5', 'TIM+R6', 'TIM+R7', 'TIM+R8',
                      'TIM+R9', 'TIM+R10', 'TIM2e', 'TIM2e+I', 'TIM2e+G',
                      'TIM2e+I+G', 'TIM2e+R2', 'TIM2e+R3', 'TIM2e+R4',
                      'TIM2e+R5', 'TIM2e+R6', 'TIM2e+R7', 'TIM2e+R8',
                      'TIM2e+R9', 'TIM2e+R10', 'TIM2', 'TIM2+I', 'TIM2+G',
                      'TIM2+I+G', 'TIM2+R2', 'TIM2+R3', 'TIM2+R4', 'TIM2+R5',
                      'TIM2+R6', 'TIM2+R7', 'TIM2+R8', 'TIM2+R9', 'TIM2+R10',
                      'TIM3e', 'TIM3e+I', 'TIM3e+G', 'TIM3e+I+G', 'TIM3e+R2',
                      'TIM3e+R3', 'TIM3e+R4', 'TIM3e+R5', 'TIM3e+R6',
                      'TIM3e+R7', 'TIM3e+R8', 'TIM3e+R9', 'TIM3e+R10', 'TIM3',
                      'TIM3+I', 'TIM3+G', 'TIM3+I+G', 'TIM3+R2', 'TIM3+R3',
                      'TIM3+R4', 'TIM3+R5', 'TIM3+R6', 'TIM3+R7', 'TIM3+R8',
                      'TIM3+R9', 'TIM3+R10', 'TVMe', 'TVMe+I', 'TVMe+G',
                      'TVMe+I+G', 'TVMe+R2', 'TVMe+R3', 'TVMe+R4', 'TVMe+R5',
                      'TVMe+R6', 'TVMe+R7', 'TVMe+R8', 'TVMe+R9', 'TVMe+R10',
                      'TVM', 'TVM+I', 'TVM+G', 'TVM+I+G', 'TVM+R2', 'TVM+R3',
                      'TVM+R4', 'TVM+R5', 'TVM+R6', 'TVM+R7', 'TVM+R8',
                      'TVM+R9', 'TVM+R10', 'SYM', 'SYM+I', 'SYM+G', 'SYM+I+G',
                      'SYM+R2', 'SYM+R3', 'SYM+R4', 'SYM+R5', 'SYM+R6',
                      'SYM+R7', 'SYM+R8', 'SYM+R9', 'SYM+R10', 'GTR', 'GTR+I',
                      'GTR+G', 'GTR+I+G', 'GTR+R2', 'GTR+R3', 'GTR+R4',
                      'GTR+R5', 'GTR+R6', 'GTR+R7', 'GTR+R8', 'GTR+R9',
                      'GTR+R10', 'MFP', 'TEST']

citations = Citations.load('citations.bib', package='q2_phylogeny')
plugin = Plugin(
    name='phylogeny',
    version=q2_phylogeny.__version__,
    website='https://github.com/qiime2/q2-phylogeny',
    package='q2_phylogeny',
    description=('This QIIME 2 plugin supports generating and manipulating '
                 'phylogenetic trees.'),
    short_description='Plugin for generating and manipulating phylogenies.'
)

plugin.methods.register_function(
    function=q2_phylogeny.midpoint_root,
    inputs={'tree': Phylogeny[Unrooted]},
    parameters={},
    outputs=[('rooted_tree', Phylogeny[Rooted])],
    input_descriptions={'tree': 'The phylogenetic tree to be rooted.'},
    parameter_descriptions={},
    output_descriptions={'rooted_tree': 'The rooted phylogenetic tree.'},
    name='Midpoint root an unrooted phylogenetic tree.',
    description=("Midpoint root an unrooted phylogenetic tree.")
)

plugin.methods.register_function(
    function=q2_phylogeny.fasttree,
    inputs={'alignment': FeatureData[AlignedSequence]},
    parameters={'n_threads': Int % Range(1, None) | Str % Choices(['auto'])},
    outputs=[('tree', Phylogeny[Unrooted])],
    input_descriptions={
        'alignment': ('Aligned sequences to be used for phylogenetic '
                      'reconstruction.')
    },
    parameter_descriptions={
        'n_threads': 'The number of threads. Using more than one thread '
                     'runs the non-deterministic variant of `FastTree` '
                     '(`FastTreeMP`), and may result in a different tree than '
                     'single-threading. See '
                     'http://www.microbesonline.org/fasttree/#OpenMP for '
                     'details. (Use `auto` to automatically use all available '
                     'cores)'
    },
    output_descriptions={'tree': 'The resulting phylogenetic tree.'},
    name='Construct a phylogenetic tree with FastTree.',
    description=("Construct a phylogenetic tree with FastTree."),
    citations=[citations['price2010fasttree']]
)

plugin.methods.register_function(
    function=q2_phylogeny.raxml,
    inputs={
            'alignment': FeatureData[AlignedSequence]},
    parameters={
            'seed': Int,
            'n_searches': Int % Range(1, None),
            'n_threads': Int % Range(1, None),
            'substitution_model': Str % Choices(_RAXML_MODEL_OPT),
            'raxml_version': Str % Choices(_RAXML_VERSION_OPT)},
    outputs=[('tree', Phylogeny[Unrooted])],
    input_descriptions={
        'alignment': ('Aligned sequences to be used for phylogenetic '
                      'reconstruction.'),
    },
    parameter_descriptions={
        'n_searches': ('The number of independent maximum likelihood '
                       'searches to perform. The single best scoring '
                       'tree is returned.'),
        'n_threads': ('The number of threads to use for multithreaded '
                      'processing. Using more than one thread '
                      'will enable the PTHREADS version of RAxML.'),
        'raxml_version': ('Select a specific CPU optimization of RAxML to '
                          'use. The SSE3 versions will run approximately 40% '
                          'faster than the standard version. The AVX2 '
                          'version will run 10-30% faster than the '
                          'SSE3 version.'),
        'substitution_model': ('Model of Nucleotide Substitution.'),
        'seed': ('Random number seed for the parsimony starting tree. '
                 'This allows you to reproduce tree results. '
                 'If not supplied then one will be randomly chosen.')},
    output_descriptions={'tree': 'The resulting phylogenetic tree.'},
    name='Construct a phylogenetic tree with RAxML.',
    description=('Construct a phylogenetic tree with RAxML. See: '
                 'https://sco.h-its.org/exelixis/web/software/raxml/'),
    citations=[citations['Stamatakis2014raxml']]
)

plugin.methods.register_function(
    function=q2_phylogeny.raxml_rapid_bootstrap,
    inputs={
            'alignment': FeatureData[AlignedSequence]},
    parameters={
            'seed': Int,
            'rapid_bootstrap_seed': Int,
            'bootstrap_replicates': Int % Range(10, None),
            'n_threads': Int % Range(1, None),
            'substitution_model': Str % Choices(_RAXML_MODEL_OPT),
            'raxml_version': Str % Choices(_RAXML_VERSION_OPT)},
    outputs=[('tree', Phylogeny[Unrooted])],
    input_descriptions={
        'alignment': ('Aligned sequences to be used for phylogenetic '
                      'reconstruction.'),
    },
    parameter_descriptions={
        'n_threads': ('The number of threads to use for multithreaded '
                      'processing. Using more than one thread '
                      'will enable the PTHREADS version of RAxML.'),
        'raxml_version': ('Select a specific CPU optimization of RAxML to '
                          'use. The SSE3 versions will run approximately 40% '
                          'faster than the standard version. The AVX2 '
                          'version will run 10-30% faster than the '
                          'SSE3 version.'),
        'substitution_model': ('Model of Nucleotide Substitution'),
        'seed': ('Random number seed for the parsimony starting tree. '
                 'This allows you to reproduce tree results. '
                 'If not supplied then one will be randomly chosen.'),
        'rapid_bootstrap_seed': ('Specify a random seed for rapid '
                                 'bootstrapping. This allows you to reproduce '
                                 'rapid bootstrap results. If not supplied '
                                 'then one will be randomly chosen.'),
        'bootstrap_replicates': ('The number of bootstrap searches to '
                                 'perform.')},
    output_descriptions={'tree': 'The resulting phylogenetic tree.'},
    name='Construct a phylogenetic tree with bootstrap supports using RAxML.',
    description=('Construct a phylogenetic tree with RAxML with the addition '
                 'of rapid bootstrapping support values. See: '
                 'https://sco.h-its.org/exelixis/web/software/raxml/'),
    citations=[citations['Stamatakis2014raxml'],
               citations['Stamatakis2008raxml']]
)

plugin.methods.register_function(
    function=q2_phylogeny.iqtree,
    inputs={'alignment': FeatureData[AlignedSequence]},
    parameters={
            'seed': Int,
            'n_cores': Int % Range(1, None) | Str % Choices(['auto']),
            'n_cores_max': Int % Range(2, None),
            'n_runs': Int % Range(1, None),
            'substitution_model': Str % Choices(_IQTREE_DNA_MODELS),
            'n_init_pars_trees': Int % Range(1, None),
            'n_top_init_trees': Int % Range(1, None),
            'n_best_retain_trees': Int % Range(1, None),
            'n_iter': Int % Range(1, None),
            'stop_iter': Int % Range(1, None),
            'perturb_nni_strength': Float % Range(0.01, 1.0),
            'spr_radius': Int % Range(1, None),
            'fast': Bool,
            'alrt': Int % Range(1000, None),
            'abayes': Bool,
            'lbp': Int % Range(1000, None),
            'allnni': Bool,
            'safe': Bool},
    outputs=[('tree', Phylogeny[Unrooted])],
    input_descriptions={
        'alignment': ('Aligned sequences to be used for phylogenetic '
                      'reconstruction.'),
    },
    parameter_descriptions={
        'n_cores': ('The number of cores to use for parallel '
                    'processing. Use `auto` to let IQ-TREE automatically '
                    'determine the optimal number of cores to use.'),
        'n_cores_max': ('Limits the maximum number of cores to be used '
                        'when \'n_cores\' is set to \'auto\'.'),
        'n_runs': ('Number of indepedent runs. Multiple  independent runs '
                   '(e.g. 10) can outperform a single run in terms of '
                   'likelihood maximisation.'),
        'substitution_model': ('Model of Nucleotide Substitution. '
                               'If not provided, IQ-TREE will determine the '
                               'best fit substitution model automatically.'),
        'seed': ('Random number seed. If not set, program defaults will be '
                 'used. See IQ-TREE manual for details.'),
        'n_init_pars_trees': ('Number of initial parsimony trees. If not '
                              'set, program defaults will be used. See '
                              'IQ-TREE manual for details.'),
        'n_top_init_trees': ('Number of top initial trees. If not set, '
                             'program defaults will be used. See IQ-TREE '
                             'manual for details.'),
        'n_best_retain_trees': ('Number of best trees retained during '
                                'search. If not set, program defaults will '
                                'be used. See IQ-TREE manual for details.'),
        'n_iter': ('Fix number of iterations to stop. If not set, program '
                   'defaults will be used. See IQ-TREE manual for details.'),
        'stop_iter': ('Number of unsuccessful iterations to stop. If not '
                      'set, program defaults will be used. See IQ-TREE '
                      'manual for details.'),
        'perturb_nni_strength': ('Perturbation strength for randomized NNI. '
                                 'If not set, program defaults will be used. '
                                 'See IQ-TREE manual for details.'),
        'spr_radius': ('Radius for parsimony SPR search. If not set, '
                       'program defaults will be used. See IQ-TREE manual '
                       'for details.'),
        'fast': ('Fast search to resemble FastTree.'),
        'allnni': ('Perform more thorough NNI search.'),
        'alrt': ('Single branch test method. Number of bootstrap replicates '
                 'to perform an SH-like approximate likelihood ratio test '
                 '(SH-aLRT). Minimum of 1000 replicates is recomended. Can '
                 'be used with other \'single branch test methods\'. Values '
                 'reported in the order of: alrt, lbp, abayes.'),
        'abayes': ('Single branch test method. Approximate Bayes test. '
                   'Can be used with other \'single branch test methods\'. '
                   'Values reported in the order of: alrt, lbp, abayes.'),
        'lbp': ('Single branch test method. Number of bootstrap replicates '
                'to perform a fast local bootstrap probability method. '
                'Minimum of 1000 replicates is recomended. Can be used with '
                'other \'single branch test methods\'. Values reported in '
                'the order of: alrt, lbp, abayes.'),
        'safe': ('Safe likelihood kernel to avoid numerical underflow.')},
    output_descriptions={'tree': 'The resulting phylogenetic tree.'},
    name='Construct a phylogenetic tree with IQ-TREE.',
    description=('Construct a phylogenetic tree using IQ-TREE '
                 '(http://www.iqtree.org/) with automatic model selection.'),
    citations=[citations['Minh2020iqtree'],
               citations['Kalyaanamoorthy2017modelfinder']],
)

plugin.methods.register_function(
    function=q2_phylogeny.iqtree_ultrafast_bootstrap,
    inputs={'alignment': FeatureData[AlignedSequence]},
    parameters={
            'seed': Int,
            'n_cores': Int % Range(1, None) | Str % Choices(['auto']),
            'n_cores_max': Int % Range(2, None),
            'n_runs': Int % Range(1, None),
            'substitution_model': Str % Choices(_IQTREE_DNA_MODELS),
            'n_init_pars_trees': Int % Range(1, None),
            'n_top_init_trees': Int % Range(1, None),
            'n_best_retain_trees': Int % Range(1, None),
            'stop_iter': Int % Range(1, None),
            'perturb_nni_strength': Float % Range(0.01, 1.0),
            'spr_radius': Int % Range(1, None),
            'bootstrap_replicates': Int % Range(1000, None),
            'n_max_ufboot_iter': Int % Range(1, None),
            'n_ufboot_steps': Int % Range(1, None),
            'min_cor_ufboot': Float % Range(0.51, 0.99),
            'ep_break_ufboot': Float % Range(0.01, 0.99),
            'alrt': Int % Range(1000, None),
            'abayes': Bool,
            'lbp': Int % Range(1000, None),
            'bnni': Bool,
            'allnni': Bool,
            'safe': Bool},
    outputs=[('tree', Phylogeny[Unrooted])],
    input_descriptions={
        'alignment': ('Aligned sequences to be used for phylogenetic '
                      'reconstruction.'),
    },
    parameter_descriptions={
        'n_cores': ('The number of cores to use for parallel '
                    'processing. Use `auto` to let IQ-TREE automatically '
                    'determine the optimal number of cores to use.'),
        'n_cores_max': ('Limits the maximum number of cores to be used '
                        'when \'n_cores\' is set to \'auto\'.'),
        'n_runs': ('Number of indepedent runs. Multiple  independent runs '
                   '(e.g. 10) can outperform a single run in terms of '
                   'likelihood maximisation.'),
        'substitution_model': ('Model of Nucleotide Substitution.'
                               'If not provided, IQ-TREE will determine the '
                               'best fit substitution model automatically. '),
        'seed': ('Random number seed. If not set, program defaults will be '
                 'used. See IQ-TREE manual for details.'),
        'bootstrap_replicates': ('The number of bootstrap searches to '
                                 'perform. Minimum of 1000 recomended. '),
        'n_init_pars_trees': ('Number of initial parsimony trees. If not '
                              'set, program defaults will be used. See '
                              'IQ-TREE manual for details.'),
        'n_top_init_trees': ('Number of top initial trees. If not set, '
                             'program defaults will be used. See IQ-TREE '
                             'manual for details.'),
        'n_best_retain_trees': ('Number of best trees retained during '
                                'search. If not set, program defaults will '
                                'be used. See IQ-TREE manual for details.'),
        'stop_iter': ('Number of unsuccessful iterations to stop. If not '
                      'set, program defaults will be used. See IQ-TREE '
                      'manual for details.'),
        'perturb_nni_strength': ('Perturbation strength for randomized NNI. '
                                 'If not set, program defaults will be used. '
                                 'See IQ-TREE manual for details.'),
        'spr_radius': ('Radius for parsimony SPR search. If not set, '
                       'program defaults will be used. See IQ-TREE manual '
                       'for details.'),
        'n_max_ufboot_iter': ('Maximum number of iterations. If not set, '
                              'program defaults will be used. See IQ-TREE '
                              'manual for details.'),
        'n_ufboot_steps': ('Number of iterations for UFBoot stopping rule. '
                           'If not set, program defaults will be used.'
                           'See IQ-TREE manual for details.'),
        'min_cor_ufboot': ('Minimum correlation coefficient. '
                           'If not set, program defaults will be used.'
                           'See IQ-TREE manual for details.'),
        'ep_break_ufboot': ('Epsilon value to break tie. If not set, program '
                            'defaults will be used. See IQ-TREE manual for '
                            'details.'),
        'bnni': ('Optimize UFBoot trees by NNI on bootstrap alignment. '
                 'This option reduces the risk of overestimating branch '
                 'supports with UFBoot due to severe model violations.'),
        'allnni': ('Perform more thorough NNI search.'),
        'alrt': ('Single branch test method. Number of bootstrap replicates '
                 'to perform an SH-like approximate likelihood ratio test '
                 '(SH-aLRT). Minimum of 1000 replicates is recomended. Can '
                 'be used with other \'single branch test methods\'. Values '
                 'reported in the order of: alrt, lbp, abayes.'),
        'abayes': ('Single branch test method. Performs an '
                   'approximate Bayes test. Can be used with other '
                   '\'single branch test methods\' and ultrafast bootstrap. '
                   'Values reported in the order of: alrt, lbp, abayes, '
                   'ufboot.'),
        'lbp': ('Single branch test method. Number of bootstrap replicates '
                'to perform a fast local bootstrap probability method. '
                'Minimum of 1000 replicates is recomended. Can be used with '
                'other \'single branch test methods\'. Values reported in '
                'the order of: alrt, lbp, abayes, ufboot.'),
        'safe': ('Safe likelihood kernel to avoid numerical underflow.')},
    output_descriptions={'tree': 'The resulting phylogenetic tree.'},
    name=('Construct a phylogenetic tree with IQ-TREE with bootstrap '
          'supports.'),
    description=('Construct a phylogenetic tree using IQ-TREE '
                 '(http://www.iqtree.org/) with automatic '
                 'model selection and bootstrap supports.'),
    citations=[citations['Minh2020iqtree'],
               citations['Kalyaanamoorthy2017modelfinder'],
               citations['Minh2013ultrafastbootstrap'],
               citations['Hoang2017ultrafastbootstrap2']],
)

T1 = TypeMatch([Frequency, RelativeFrequency, PresenceAbsence])

plugin.methods.register_function(
    function=q2_phylogeny.filter_table,
    inputs={'table': FeatureTable[T1],
            'tree': Phylogeny[Rooted | Unrooted]},
    parameters={},
    outputs=[('filtered_table', FeatureTable[T1])],
    input_descriptions={
        'table': 'Feature table that features should be filtered from.',
        'tree': ('Tree where tip identifiers are the feature identifiers that '
                 'should be retained in the table.')
    },
    parameter_descriptions={},
    output_descriptions={'filtered_table': 'The resulting feature table.'},
    name="Remove features from table if they're not present in tree.",
    description=("Remove features from a feature table if their identifiers "
                 "are not tip identifiers in tree.")
)

T2 = TypeMatch([Rooted, Unrooted])

T3 = (Frequency | RelativeFrequency | PresenceAbsence | Composition)

plugin.methods.register_function(
    function=q2_phylogeny.filter_tree,
    inputs={'tree': Phylogeny[T2],
            'table': FeatureTable[T3],
            },
    parameters={'metadata': Metadata,
                'where': Str
                },
    outputs=[('filtered_tree', Phylogeny[T2])],
    input_descriptions={
        'tree': ('Tree that should be filtered'),
        'table': ('Feature table which contains the identifier that should be'
                  ' retained in the tree'),
    },
    parameter_descriptions={
        'metadata': ("Feature metadata to use with the 'where' statement or "
                     "to select tips to be retained. Metadata objects could "
                     "also include FeatureData[Sequence] data types, if, for"
                     "instance, you want to filter to match represenative "
                     "sequencces."),
        'where': ('SQLite WHERE clause specifying sample metadata criteria '
                  'that must be met to be included in the filtered feature '
                  'table. If not provided, all samples in `metadata` that'
                  ' are also in the feature table will be retained.'),
    },
    output_descriptions={'filtered_tree': 'The resulting phylogenetic tree.'},
    name="Remove features from tree based on metadata",
    description=("Remove tips from a tree if their identifiers based on a "
                 "set of provided identifiers.")
)

plugin.methods.register_function(
    function=q2_phylogeny.robinson_foulds,
    inputs={'trees': List[Phylogeny[Rooted | Unrooted]]},
    parameters={
        'labels': List[Str],
        'missing_tips': Str % Choices('error', 'intersect-all')
    },
    outputs=[('distance_matrix', DistanceMatrix)],
    input_descriptions={
        'trees': 'Phylogenetic trees to compare with Robinson-Foulds. Rooting'
                 ' information and branch lengths are ignored by this metric.'
    },
    parameter_descriptions={
        'labels': 'Labels to use for the tree names in the distance matrix.'
                  ' If ommited, labels will be "tree_n" where "n" ranges from'
                  ' 1..N. The number of labels must match the number of'
                  ' trees.',
        'missing_tips': 'How to handle tips that are not shared between trees.'
                        ' "error" will raise an error if the set of tips is'
                        ' not identical between all input trees.'
                        ' "intersect-all" will remove tips that are not shared'
                        ' between all trees before computing distances beteen'
                        ' trees.'
    },
    output_descriptions={
        'distance_matrix': 'The distances between trees as a symmetric matrix.'
    },
    name="Calculate Robinson-Foulds distance between phylogenetic trees.",
    description="Calculate the Robinson-Foulds symmetric difference metric"
                " between two or more phylogenetic trees.",
    citations=[citations['robinson1981comparison']]
)

plugin.pipelines.register_function(
    function=q2_phylogeny.align_to_tree_mafft_fasttree,
    inputs={
        'sequences': FeatureData[Sequence],
    },
    parameters={
        'n_threads': Int % Range(1, None) | Str % Choices(['auto']),
        'mask_max_gap_frequency': Float % Range(0, 1, inclusive_end=True),
        'mask_min_conservation': Float % Range(0, 1, inclusive_end=True),
        'parttree': Bool,
    },
    outputs=[
        ('alignment', FeatureData[AlignedSequence]),
        ('masked_alignment', FeatureData[AlignedSequence]),
        ('tree', Phylogeny[Unrooted]),
        ('rooted_tree', Phylogeny[Rooted]),
    ],
    input_descriptions={
        'sequences': 'The sequences to be used for creating a '
                     'fasttree based rooted phylogenetic tree.'
    },
    parameter_descriptions={
        'n_threads': 'The number of threads. (Use `auto` to automatically use '
                     'all available cores) '
                     'This value is used when aligning the sequences and '
                     'creating the tree with fasttree.',
        'mask_max_gap_frequency': 'The maximum relative frequency of gap '
                                  'characters in a column for the column '
                                  'to be retained. This relative frequency '
                                  'must be a number between 0.0 and 1.0 '
                                  '(inclusive), where 0.0 retains only those '
                                  'columns without gap characters, and 1.0 '
                                  'retains all columns  regardless of gap '
                                  'character frequency. This value is used '
                                  'when masking the aligned sequences.',
        'mask_min_conservation':  'The minimum relative frequency '
                                  'of at least one non-gap character in a '
                                  'column for that column to be retained. '
                                  'This relative frequency must be a number '
                                  'between 0.0 and 1.0 (inclusive). For '
                                  'example, if a value of  0.4 is provided, a '
                                  'column will only be retained  if it '
                                  'contains at least one character that is '
                                  'present in at least 40% of the sequences. '
                                  'This value is used when masking the '
                                  'aligned sequences.',
        'parttree': 'This flag is required if the number of sequences being '
                    'aligned are larger than 1000000. Disabled by default.',
    },
    output_descriptions={
        'alignment': 'The aligned sequences.',
        'masked_alignment': 'The masked alignment.',
        'tree': 'The unrooted phylogenetic tree.',
        'rooted_tree': 'The rooted phylogenetic tree.',
    },
    name='Build a phylogenetic tree using fasttree and mafft alignment',
    description=('This pipeline will start by creating a sequence alignment '
                 'using MAFFT, after which any alignment columns that are '
                 'phylogenetically uninformative or ambiguously aligned will '
                 'be removed (masked). The resulting masked alignment will be '
                 'used to infer a phylogenetic tree and then subsequently '
                 'rooted at its midpoint. Output files from each step of the '
                 'pipeline will be saved. This includes both the unmasked and '
                 'masked MAFFT alignment from q2-alignment methods, and both '
                 'the rooted and unrooted phylogenies from q2-phylogeny '
                 'methods.'
                 ),
    examples={
        'align_to_tree_mafft_fasttree':
        ex.phylogeny_align_to_tree_mafft_fasttree,
    }
)

plugin.pipelines.register_function(
    function=q2_phylogeny.align_to_tree_mafft_iqtree,
    inputs={
        'sequences': FeatureData[Sequence],
    },
    parameters={
        'n_threads': Int % Range(1, None) | Str % Choices(['auto']),
        'mask_max_gap_frequency': Float % Range(0, 1, inclusive_end=True),
        'mask_min_conservation': Float % Range(0, 1, inclusive_end=True),
        'seed': Int,
        'substitution_model': Str % Choices(_IQTREE_DNA_MODELS),
        'stop_iter': Int % Range(1, None),
        'perturb_nni_strength': Float % Range(0.01, 1.0),
        'fast': Bool,
        'alrt': Int % Range(1000, None),
    },
    outputs=[
        ('alignment', FeatureData[AlignedSequence]),
        ('masked_alignment', FeatureData[AlignedSequence]),
        ('tree', Phylogeny[Unrooted]),
        ('rooted_tree', Phylogeny[Rooted]),
    ],
    input_descriptions={
        'sequences': 'The sequences to be used for creating a '
                     'iqtree based rooted phylogenetic tree.'
    },
    parameter_descriptions={
        'n_threads': 'The number of threads. (Use 0 to automatically use all '
                     'available cores '
                     'This value is used when aligning the sequences and '
                     'creating the tree with iqtree.',
        'mask_max_gap_frequency': 'The maximum relative frequency of gap '
                                  'characters in a column for the column '
                                  'to be retained. This relative frequency '
                                  'must be a number between 0.0 and 1.0 '
                                  '(inclusive), where 0.0 retains only those '
                                  'columns without gap characters, and 1.0 '
                                  'retains all columns  regardless of gap '
                                  'character frequency. This value is used '
                                  'when masking the aligned sequences.',
        'mask_min_conservation':  'The minimum relative frequency '
                                  'of at least one non-gap character in a '
                                  'column for that column to be retained. '
                                  'This relative frequency must be a number '
                                  'between 0.0 and 1.0 (inclusive). For '
                                  'example, if a value of  0.4 is provided, a '
                                  'column will only be retained  if it '
                                  'contains at least one character that is '
                                  'present in at least 40% of the sequences. '
                                  'This value is used when masking the '
                                  'aligned sequences.',
        'seed':  'Random number seed for the iqtree parsimony starting tree. '
                 'This allows you to reproduce tree results. '
                 'If not supplied then one will be randomly chosen.',

        'substitution_model':  'Model of Nucleotide Substitution. '
                               'If not provided, IQ-TREE will determine the '
                               'best fit substitution model automatically.',
        'stop_iter': 'Number of unsuccessful iterations to stop. If not '
                     'set, program defaults will be used. See IQ-TREE '
                     'manual for details.',
        'perturb_nni_strength': 'Perturbation strength for randomized NNI. '
                                'If not set, program defaults will be used. '
                                'See IQ-TREE manual for details.',
        'fast': 'Fast search to resemble FastTree.',
        'alrt': 'Single branch test method. Number of bootstrap replicates '
                 'to perform an SH-like approximate likelihood ratio test '
                 '(SH-aLRT). Minimum of 1000 replicates is recomended.'
    },
    output_descriptions={
        'alignment': 'The aligned sequences.',
        'masked_alignment': 'The masked alignment.',
        'tree': 'The unrooted phylogenetic tree.',
        'rooted_tree': 'The rooted phylogenetic tree.',
    },
    name='Build a phylogenetic tree using iqtree and mafft alignment.',
    description=('This pipeline will start by creating a sequence alignment '
                 'using MAFFT, after which any alignment columns that are '
                 'phylogenetically uninformative or ambiguously aligned will '
                 'be removed (masked). The resulting masked alignment will be '
                 'used to infer a phylogenetic tree using IQ-TREE. By default '
                 'the best fit substitution model will be determined by '
                 'ModelFinder prior to phylogenetic inference. The resulting '
                 'tree will be subsequently rooted at its midpoint. Output '
                 'files from each step of the pipeline will be saved. This '
                 'includes both the unmasked and masked MAFFT alignment from '
                 'q2-alignment methods, and both the rooted and unrooted '
                 'phylogenies from q2-phylogeny methods.'
                 ),
)

plugin.pipelines.register_function(
    function=q2_phylogeny.align_to_tree_mafft_raxml,
    inputs={
        'sequences': FeatureData[Sequence],
    },
    parameters={
        'n_threads': Int % Range(1, None) | Str % Choices(['auto']),
        'mask_max_gap_frequency': Float % Range(0, 1, inclusive_end=True),
        'mask_min_conservation': Float % Range(0, 1, inclusive_end=True),
        'parttree': Bool,
        'seed': Int,
        'substitution_model': Str % Choices(_RAXML_MODEL_OPT),
        'raxml_version': Str % Choices(_RAXML_VERSION_OPT),
    },
    outputs=[
        ('alignment', FeatureData[AlignedSequence]),
        ('masked_alignment', FeatureData[AlignedSequence]),
        ('tree', Phylogeny[Unrooted]),
        ('rooted_tree', Phylogeny[Rooted]),
    ],
    input_descriptions={
        'sequences': 'The sequences to be used for creating a '
                     'iqtree based rooted phylogenetic tree.'
    },
    parameter_descriptions={
        'n_threads': 'The number of threads. (Use `all` to automatically use '
                     'all available cores. This value is used when aligning '
                     'the sequences and creating the tree with iqtree.',
        'mask_max_gap_frequency': 'The maximum relative frequency of gap '
                                  'characters in a column for the column '
                                  'to be retained. This relative frequency '
                                  'must be a number between 0.0 and 1.0 '
                                  '(inclusive), where 0.0 retains only those '
                                  'columns without gap characters, and 1.0 '
                                  'retains all columns  regardless of gap '
                                  'character frequency. This value is used '
                                  'when masking the aligned sequences.',
        'mask_min_conservation':  'The minimum relative frequency '
                                  'of at least one non-gap character in a '
                                  'column for that column to be retained. '
                                  'This relative frequency must be a number '
                                  'between 0.0 and 1.0 (inclusive). For '
                                  'example, if a value of  0.4 is provided, a '
                                  'column will only be retained  if it '
                                  'contains at least one character that is '
                                  'present in at least 40% of the sequences. '
                                  'This value is used when masking the '
                                  'aligned sequences.',
        'parttree': 'This flag is required if the number of sequences being '
                    'aligned are larger than 1000000. Disabled by default. '
                    'NOTE: if using this option, it is recomended that only '
                    'the CAT-based substitution models of RAxML be '
                    'considered for this pipeline.',
        'seed':  'Random number seed for the parsimony starting tree. '
                 'This allows you to reproduce tree results. '
                 'If not supplied then one will be randomly chosen.',
        'raxml_version': ('Select a specific CPU optimization of RAxML to '
                          'use. The SSE3 versions will run approximately 40% '
                          'faster than the standard version. The AVX2 '
                          'version will run 10-30% faster than the '
                          'SSE3 version.'),
        'substitution_model': ('Model of Nucleotide Substitution.'),
    },
    output_descriptions={
        'alignment': 'The aligned sequences.',
        'masked_alignment': 'The masked alignment.',
        'tree': 'The unrooted phylogenetic tree.',
        'rooted_tree': 'The rooted phylogenetic tree.',
    },
    name='Build a phylogenetic tree using raxml and mafft alignment.',
    description=('This pipeline will start by creating a sequence alignment '
                 'using MAFFT, after which any alignment columns that are '
                 'phylogenetically uninformative or ambiguously aligned will '
                 'be removed (masked). The resulting masked alignment will be '
                 'used to infer a phylogenetic tree using RAxML, under the '
                 'specified substitution model, and then subsequently '
                 'rooted at its midpoint. Output files from each step of the '
                 'pipeline will be saved. This includes both the unmasked and '
                 'masked MAFFT alignment from q2-alignment methods, and both '
                 'the rooted and unrooted phylogenies from q2-phylogeny '
                 'methods.'
                 )
)



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2022, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import unittest
import io

import skbio

from q2_phylogeny import midpoint_root, robinson_foulds


class MidpointRootTests(unittest.TestCase):

    def test_midpoint_root(self):
        unrooted_nwk = io.StringIO("((a:1,b:1):1,(c:5,d:4):1);")
        rooted_nwk = io.StringIO("(c:4.5,(d:4,(a:1,b:1):2):0.5);")
        unrooted_tree = skbio.TreeNode.read(unrooted_nwk)
        expected = skbio.TreeNode.read(rooted_nwk)
        actual = midpoint_root(unrooted_tree)
        # skbio.TreeNode currently does not have an equality operator
        # (see https://github.com/biocore/scikit-bio/issues/465), so do a
        # few specific comparisons of actual and expected

        # confirm ids are as expected
        actual_tip_tip_distances = actual.tip_tip_distances()
        expected_tip_tip_distances = expected.tip_tip_distances()
        self.assertEqual(sorted(actual_tip_tip_distances.ids),
                         sorted(expected_tip_tip_distances.ids))

        # confirm tip-to-tip distances are as expected
        for id1 in actual_tip_tip_distances.ids:
            for id2 in actual_tip_tip_distances.ids:
                actual_distance = actual_tip_tip_distances[id1, id2]
                expected_distance = expected_tip_tip_distances[id1, id2]
                self.assertEqual(
                 actual_distance, expected_distance,
                 "Distances between %s and %s are not equal. (%f versus %f)." %
                 (id1, id2, actual_distance, expected_distance))

        # confirm tip-to-root distances are as expected
        for id_ in actual_tip_tip_distances.ids:
            self.assertEqual(actual.find(id_).distance(actual.root()),
                             expected.find(id_).distance(expected.root()))


class TestRobinsonFoulds(unittest.TestCase):
    def test_expected(self):
        # Using Felsenstein's test data set from here:
        # http://evolution.genetics.washington.edu/phylip/doc/treedist.html
        test_data_set_1 = [  # name from original source
            ['(A,(B,(H,(D,(J,(((G,E),(F,I)),C))))));'],
            ['(A,(B,(D,((J,H),(((G,E),(F,I)),C)))));'],
            ['(A,(B,(D,(H,(J,(((G,E),(F,I)),C))))));'],
            ['(A,(B,(E,(G,((F,I),((J,(H,D)),C))))));'],
            ['(A,(B,(E,(G,((F,I),(((J,H),D),C))))));'],
            ['(A,(B,(E,((F,I),(G,((J,(H,D)),C))))));'],
            ['(A,(B,(E,((F,I),(G,(((J,H),D),C))))));'],
            ['(A,(B,(E,((G,(F,I)),((J,(H,D)),C)))));'],
            ['(A,(B,(E,((G,(F,I)),(((J,H),D),C)))));'],
            ['(A,(B,(E,(G,((F,I),((J,(H,D)),C))))));'],
            ['(A,(B,(D,(H,(J,(((G,E),(F,I)),C))))));'],
            ['(A,(B,(E,((G,(F,I)),((J,(H,D)),C)))));']
        ]
        trees = [skbio.TreeNode.read(nwk) for nwk in test_data_set_1]
        expected = skbio.DistanceMatrix(
            [
                [0,   4,  2, 10, 10, 10, 10, 10, 10, 10,  2, 10],
                [4,   0,  2, 10,  8, 10,  8, 10,  8, 10,  2, 10],
                [2,   2,  0, 10, 10, 10, 10, 10, 10, 10,  0, 10],
                [10, 10, 10,  0,  2,  2,  4,  2,  4,  0, 10,  2],
                [10,  8, 10,  2,  0,  4,  2,  4,  2,  2, 10,  4],
                [10, 10, 10,  2,  4,  0,  2,  2,  4,  2, 10,  2],
                [10,  8, 10,  4,  2,  2,  0,  4,  2,  4, 10,  4],
                [10, 10, 10,  2,  4,  2,  4,  0,  2,  2, 10,  0],
                [10,  8, 10,  4,  2,  4,  2,  2,  0,  4, 10,  2],
                [10, 10, 10,  0,  2,  2,  4,  2,  4,  0, 10,  2],
                [2,   2,  0, 10, 10, 10, 10, 10, 10, 10,  0, 10],
                [10, 10, 10,  2,  4,  2,  4,  0,  2,  2, 10,  0]
            ], ids=['tree_%d' % d for d in range(1, 13)])

        result = robinson_foulds(trees)

        self.assertEqual(result, expected)

    def test_single_tree_and_label(self):
        trees = [skbio.TreeNode.read(['(A:0.2, B:1.5, C, (E, F));'])]
        expected = skbio.DistanceMatrix([[0]], ids=['foo'])

        result = robinson_foulds(trees, labels=['foo'])

        self.assertEqual(result, expected)

    def test_labels_too_short(self):
        test_data = [
            ['(A,(B,(H,(D,(J,(((G,E),(F,I)),C))))));'],
            ['(A,(B,(D,((J,H),(((G,E),(F,I)),C)))));'],
        ]
        trees = [skbio.TreeNode.read(nwk) for nwk in test_data]

        with self.assertRaisesRegex(ValueError, 'number.*match'):
            robinson_foulds(trees, labels=['A'])

    def test_missing_single_tip(self):
        test_data = [
            ['(A,(B,(H,(D,(J,(((G,E),(F,I)),C))))));'],
            ['((B,(D,((J,H),(((G,E),(F,I)),C)))));'],
        ]
        trees = [skbio.TreeNode.read(nwk) for nwk in test_data]

        with self.assertRaisesRegex(ValueError, "tips.*shared.*'A'"):
            robinson_foulds(trees)

    def test_missing_many_tips(self):
        test_data = [
            ['(A,(B,(H,(D,(J,(((G,E),(F,I,K,L)),C))))));'],
            ['(A,(,((,),(((,),(,)),))));'],
        ]
        trees = [skbio.TreeNode.read(nwk) for nwk in test_data]

        with self.assertRaisesRegex(ValueError, "tips.*shared"):
            robinson_foulds(trees)

    def test_missing_all_tips(self):
        test_data = [
            ['(A,(B,(H,(D,(J,(((G,E),(F,I,K,L)),C))))));'],
            ['(,(,((,),(((,),(,)),))));'],
        ]
        trees = [skbio.TreeNode.read(nwk) for nwk in test_data]

        with self.assertRaisesRegex(ValueError, "No tip names.*shared"):
            robinson_foulds(trees)

    def test_missing_intersect_all(self):
        test_data = [
            ['(A,(B,(H,(J,(((G,E),(F,I)),C)))));'],
            ['(B,((J,H),(((G,E),(F,I)),C)));'],
            ['(B,(D,(H,(J,(((G,E),(F,I)),C)))));'],
        ]
        trees = [skbio.TreeNode.read(nwk) for nwk in test_data]
        expected = skbio.DistanceMatrix([[0, 2, 0], [2, 0, 2], [0, 2, 0]])

        result = robinson_foulds(trees, labels=['0', '1', '2'],
                                 missing_tips='intersect-all')

        self.assertEqual(result, expected)

    def test_invalid_missing(self):
        test_data = [
            ['(A,(B,(H,(D,(J,(((G,E),(F,I,K,L)),C))))));'],
        ]
        trees = [skbio.TreeNode.read(nwk) for nwk in test_data]

        with self.assertRaisesRegex(ValueError, "not-an-option"):
            robinson_foulds(trees, missing_tips="not-an-option")


if __name__ == "__main__":
    unittest.main()



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2022, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------


rep_seqs_url = ('https://data.qiime2.org/usage-examples/'
                'moving-pictures/rep-seqs-dada2.qza')


def phylogeny_align_to_tree_mafft_fasttree(use):
    rep_seqs = use.init_artifact_from_url('rep_seqs', rep_seqs_url)

    alignment, masked_alignment, tree, rooted_tree = use.action(
        use.UsageAction('phylogeny', 'align_to_tree_mafft_fasttree'),
        use.UsageInputs(sequences=rep_seqs),
        use.UsageOutputNames(
            alignment='aligned-rep-seqs',
            masked_alignment='masked-aligned-rep-seqs',
            tree='unrooted-tree',
            rooted_tree='rooted-tree',
        )
    )

    alignment.assert_output_type('FeatureData[AlignedSequence]')
    masked_alignment.assert_output_type('FeatureData[AlignedSequence]')
    tree.assert_output_type('Phylogeny[Unrooted]')
    rooted_tree.assert_output_type('Phylogeny[Rooted]')



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2022, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import unittest
import io

import numpy as np
import pandas as pd
from biom.table import Table
import skbio

from qiime2 import Metadata

from q2_phylogeny import (filter_table, filter_tree)


class FilterTableTests(unittest.TestCase):

    def test_tree_filter_table_some(self):
        rooted_nwk = io.StringIO("(O1:4.5,(d:4,(a:1,b:1):2):0.5);")
        tree = skbio.TreeNode.read(rooted_nwk)
        table = Table(np.array([[0, 1, 3], [1, 1, 2]]),
                      ['O1', 'O2'],
                      ['S1', 'S2', 'S3'])
        actual = filter_table(table, tree)
        expected = table.filter(['O1'], axis='observation')
        self.assertEqual(actual, expected)

    def test_tree_filter_table_all(self):
        rooted_nwk = io.StringIO("(c:4.5,(d:4,(a:1,b:1):2):0.5);")
        tree = skbio.TreeNode.read(rooted_nwk)
        table = Table(np.array([[0, 1, 3], [1, 1, 2]]),
                      ['O1', 'O2'],
                      ['S1', 'S2', 'S3'])
        actual = filter_table(table, tree)
        expected = table.filter([], axis='observation')
        self.assertEqual(actual, expected)

    def test_tree_filter_table_none(self):
        rooted_nwk = io.StringIO("(O1:4.5,(O2:4,(a:1,b:1):2):0.5);")
        tree = skbio.TreeNode.read(rooted_nwk)
        table = Table(np.array([[0, 1, 3], [1, 1, 2]]),
                      ['O1', 'O2'],
                      ['S1', 'S2', 'S3'])
        actual = filter_table(table, tree)
        expected = table.filter(['O1', 'O2'], axis='observation')
        self.assertEqual(actual, expected)


class FilterTreeTests(unittest.TestCase):
    def setUp(self):
        rooted_nwk = io.StringIO("((A:0.1, B:0.2)C:0.3, D:0.4, E:0.5)root;")
        self.tree = skbio.TreeNode.read(rooted_nwk)
        self.metadata = Metadata(pd.DataFrame(
                data=np.array([['Bacteria', '1'],
                               ['Archea', '1']], dtype=object),
                index=pd.Index(['A', 'D'], name='Feature ID'),
                columns=['kingdom', 'keep'],
            ))
        self.table = Table(data=np.array([[0, 1, 2], [2, 2, 2]]),
                           observation_ids=['A', 'D'],
                           sample_ids=['S1', 'S2', 'S3']
                           )
        self.filtered_tree = self.tree.copy().shear(['A', 'D'])
        self.filtered_tree.prune()

    def test_filter_tree_error_no_filter_art(self):
        with self.assertRaises(ValueError) as err:
            filter_tree(self.tree)
        self.assertEqual(
            str(err.exception),
            ('A feature table, sequences or metadata must be provided for '
                'filtering.')
            )

    def test_filter_tree_error_multiple_filter_arts(self):
        with self.assertRaises(ValueError) as err:
            filter_tree(self.tree,
                        table=self.table,
                        metadata=self.metadata)
        self.assertEqual(
            str(err.exception),
            ('Filtering can only be performed using one reference file. '
             'Please choose between filtering with a feature table, '
             'sequences, or metadata.')
        )

    def test_filter_tree_error_where_no_metadata(self):
        with self.assertRaises(ValueError) as err:
            filter_tree(self.tree,
                        table=self.table,
                        where='[kingdom]="Archea"')
        self.assertEqual(
            str(err.exception),
            ("Metadata must be provided if 'where' is specified")
            )

    def test_filter_tree_error_filter_superset(self):
        metadata = Metadata(pd.DataFrame(
            data=np.array([[1, 1, 0]]).T,
            index=pd.Index(['A', 'D', 'F'], name='feature-id'),
            columns=['keep']
        ))
        with self.assertRaises(ValueError) as err:
            filter_tree(self.tree,
                        metadata=metadata)
        self.assertEqual(
            str(err.exception),
            'The ids for filtering must be a subset of the tips in the tree.'
            )

    def test_filter_tree_metadata(self):
        test = filter_tree(self.tree, metadata=self.metadata)
        self.assertEqual(str(test), str(self.filtered_tree))

    def test_filter_tree_table(self):
        test = filter_tree(self.tree, table=self.table)
        self.assertEqual(str(test), str(self.filtered_tree))


if __name__ == "__main__":
    unittest.main()
